#include <iostream>
#include <algorithm>
#include <unordered_set>
#include <set>
#include "MeshLoader.h"
#include "Hash.h"
#include <fstream>
#include <ranges>


const std::unordered_map<int, Node>& MeshLoader::getNodes() {
    return Nodes;
}


const std::unordered_map<int, Tetrahedron>& MeshLoader::getFiniteElements() {
    return FiniteElements;
}


const std::unordered_map<int, Triangle>& MeshLoader::getBoundaryElements() {
    return BoundaryElements;
}


void MeshLoader::printNodes(const std::unordered_map<int, Node>& p_nodes) {
    std::cout << p_nodes;
}


void MeshLoader::printFiniteElements(const std::unordered_map<int, Tetrahedron>& p_elem) {
    std::cout << p_elem;
}


void MeshLoader::printBoundaryElements(const std::unordered_map<int, Triangle>& p_elem) {
    std::cout << p_elem;
}


// ID    
std::unordered_set<int> MeshLoader::getFiniteElementIDByNodesID(int p_id1, int p_id2, int p_id3) {
    std::unordered_set<int> res;

    for (const auto& pair : FiniteElements) {
        const Tetrahedron& elem = pair.second;
        if (std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id1) != elem.m_NodeIDs.end() &&
            std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id2) != elem.m_NodeIDs.end() &&
            std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id3) != elem.m_NodeIDs.end()) {
            res.insert(pair.first);
        }
    }
    return res;
}


// ID    ()
std::unordered_set<int> MeshLoader::getFiniteElementIDByEdge(int p_id1, int p_id2) {
    std::unordered_set<int> res;
    for (const auto& elementPair : FiniteElements) {
        const Tetrahedron& element = elementPair.second;
        if (std::find(element.m_NodeIDs.begin(), element.m_NodeIDs.end(), p_id1) != element.m_NodeIDs.end() &&
            std::find(element.m_NodeIDs.begin(), element.m_NodeIDs.end(), p_id2) != element.m_NodeIDs.end()) {
            res.insert(elementPair.first);
        }
    }
    return res;
}


//  ID   ID 
std::unordered_set<int> MeshLoader::getNodesIDByBorder(int p_id) {
    std::unordered_set<int> res;
    for (const auto& elementPair : BoundaryElements) {
        const Triangle& elem = elementPair.second;
        if (elem.m_GeometryAreaID == p_id) {
            for (const auto& nodeId : elem.m_NodeIDs) {
                res.insert(nodeId);
            }
        }
    }
    return res;
}


//  ID   ID 
std::unordered_set<int> MeshLoader::getFiniteElementIdBySurface(int p_id) {
    std::unordered_set<int> res;
    for (const auto& elementPair : FiniteElements) {
        const int elementId = elementPair.first;
        const Tetrahedron& element = elementPair.second;
        if (element.m_GeometryAreaID == p_id) {
            res.insert(elementId);
        }
    }
    return res;
}


//  ID    ID 
std::unordered_set<int> MeshLoader::getBoundElementIDByBorder(int _id) {
    std::unordered_set<int> res;
    for (const auto& pair : BoundaryElements) {
        const int elementId = pair.first;
        const Triangle& elem = pair.second;
        if (elem.m_GeometryAreaID == _id) {
            res.insert(elementId);
        }
    }
    return res;
}


//    
std::vector<Tetrahedron> MeshLoader::getFiniteElementByNodesID(int p_id1, int p_id2, int p_id3) {
    std::vector<Tetrahedron> res;

    for (const auto& pair : FiniteElements) {
        const Tetrahedron& elem = pair.second;
        if (std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id1) != elem.m_NodeIDs.end() &&
            std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id2) != elem.m_NodeIDs.end() &&
            std::find(elem.m_NodeIDs.begin(), elem.m_NodeIDs.end(), p_id3) != elem.m_NodeIDs.end()) {
            res.push_back(pair.second);
        }
    }
    return res;
}


//    ()
std::vector<Tetrahedron> MeshLoader::getFiniteElementByEdge(int p_id1, int p_id2) {
    std::vector<Tetrahedron> res;
    for (const auto& elementPair : FiniteElements) {
        const Tetrahedron& element = elementPair.second;
        if (std::find(element.m_NodeIDs.begin(), element.m_NodeIDs.end(), p_id1) != element.m_NodeIDs.end() &&
            std::find(element.m_NodeIDs.begin(), element.m_NodeIDs.end(), p_id2) != element.m_NodeIDs.end()) {
            res.push_back(elementPair.second);
        }
    }
    return res;
}


//    ID 
std::vector<Node> MeshLoader::getNodesByBorder(int p_id) {
    std::vector<Node> res;
    for (const auto& elementPair : BoundaryElements) {
        const Triangle& elem = elementPair.second;
        if (elem.m_GeometryAreaID == p_id) {
            for (const auto& nodeId : elem.m_NodeIDs) {
                res.push_back(Nodes[nodeId]);
            }
        }
    }
    return res;
}


//    ID 
std::vector<Tetrahedron> MeshLoader::getFiniteElementBySurface(int p_id) {
    std::vector<Tetrahedron> res;
    for (const auto& elementPair : FiniteElements) {
        const Tetrahedron& elem = elementPair.second;
        if (elem.m_GeometryAreaID == p_id) {
            res.push_back(elem);
        }
    }
    return res;
}


//     ID 
std::vector<Triangle> MeshLoader::getBoundElementByBorder(int _id) {
    std::vector<Triangle> res;
    for (const auto& pair : BoundaryElements) {
        const Triangle& elem = pair.second;
        if (elem.m_GeometryAreaID == _id) {
            res.push_back(elem);
        }
    }
    return res;
}


//  ID  
std::unordered_map<int, std::set<int>> MeshLoader::createNeighborsMap() {
    std::unordered_map<int, std::set<int>> neighbors;
    for (const auto& pair : BoundaryElements) {
        const Triangle& elem = pair.second;
        for (auto nodeId : elem.m_NodeIDs) {
            for (auto anthNodeId : elem.m_NodeIDs) {
                if (nodeId != anthNodeId) {
                    neighbors[nodeId].insert(anthNodeId);
                }
            }
        }
    }

    for (const auto& pair : neighbors) {
        std::cout << "Neighbours of node " << pair.first << ": ";
        for (const auto& neighborId : pair.second) {
            std::cout << neighborId << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;

    return neighbors;
}


//     
Node MeshLoader::getMiddleNode(const Edge& edge) {
    double MiddleX = (Nodes[edge.m_edgeNodes.first].m_coordinates[0] +
        Nodes[edge.m_edgeNodes.second].m_coordinates[0]) / 2.0;
    double MiddleY = (Nodes[edge.m_edgeNodes.first].m_coordinates[1] +
        Nodes[edge.m_edgeNodes.second].m_coordinates[1]) / 2.0;
    double MiddleZ = (Nodes[edge.m_edgeNodes.first].m_coordinates[2] +
        Nodes[edge.m_edgeNodes.second].m_coordinates[2]) / 2.0;
    Node newNode({ MiddleX, MiddleY, MiddleZ }, (Nodes[edge.m_edgeNodes.first].m_vertexDot && Nodes[edge.m_edgeNodes.first].m_vertexDot));
    return newNode;
}


void MeshLoader::insertNodeMid() {
    std::unordered_set<Edge, Hash> edges;
    std::unordered_map<Edge, int, Hash> edgeToMidNodeId;

    for (auto& elemPair : FiniteElements) {
        Tetrahedron& elem = elemPair.second;
        std::vector<int> newMidNodeIDs;
        for (int i = 0; i < elem.m_NodeIDs.size(); ++i) {
            for (int j = i + 1; j < elem.m_NodeIDs.size(); ++j) {
                Edge curEdge(elem.m_NodeIDs[i], elem.m_NodeIDs[j]);
                if (edges.insert(curEdge).second) {
                    Node newNode = getMiddleNode(curEdge);
                    int newId = m_nodeId.generateId();
                    Nodes[newId] = newNode;
                    edgeToMidNodeId[curEdge] = newId;
                    newMidNodeIDs.push_back(newId);
                }
            }
        }
        elem.m_NodeIDs.insert(elem.m_NodeIDs.end(), newMidNodeIDs.begin(), newMidNodeIDs.end());
    }
    for (auto& belemPair : BoundaryElements) {
        Triangle& belem = belemPair.second;
        for (int i = 0; i < belem.m_NodeIDs.size(); ++i) {
            for (int j = i + 1; j < belem.m_NodeIDs.size(); ++j) {
                Edge curEdge(belem.m_NodeIDs[i], belem.m_NodeIDs[j]);
                auto it = edgeToMidNodeId.find(curEdge);
                if (it == edgeToMidNodeId.end()) {
                    curEdge = { curEdge.m_edgeNodes.second, curEdge.m_edgeNodes.first };
                    it = edgeToMidNodeId.find(curEdge);
                }
                if (it != edgeToMidNodeId.end()) {
                    belem.m_NodeIDs.push_back(it->second);
                }
            }
        }
    }
}
